use hyper::upgrade::Upgraded;
use hyper_util::rt::TokioIo;
use tokio_tungstenite::WebSocketStream;
use futures::{SinkExt, StreamExt};

use crate::{
    protocol,
    util::pack::{pack_u16, unpack_u16},
    HttpServiceArgs,
};

fn encode_response(code:u16, data:Vec<u8>) -> Vec<u8>
{
    [
        pack_u16(code),
        data,
    ].concat()
}

pub async fn handle_ws(mut ws:WebSocketStream<TokioIo<Upgraded>>, args:HttpServiceArgs) -> Result<(),()>
// Handle websocket connection.
//
{
    use tokio_tungstenite::tungstenite::protocol::Message;
    use protocol::{QRPacket, QRPacketData::*, code::*, packet::*};

    let conn_id :u32;

    let bus_ds = args.bus.mailbox(1).unwrap_or(1);

    // Perform connection handshake with data system.
    //  - Provide system with connection/bus pairing.
    //  - Acquire connection id.
    //
    args.bus.send(bus_ds, QRPacket::new(0, QConn(LocalPacketConnect {
        bus_id:args.bus.id(),
    })))?;
    match args.bus.receive_wait() {
        Some(resp) => {
            let qr = &resp.data;
            match qr.data {
                RConn => { conn_id = qr.id; }
                _ => { return Err(()); }
            }
        }
        None => { return Err(()); }
    }

    // Decode client requests from websocket,
    // pass requests to data system,
    // and return responses to client.
    while match ws.next().await {
        Some(msg) => match msg {
            Ok(msg) => match msg {
                Message::Binary(data) => {
                    let mut index :usize = 0;
                    let code: u16 = unpack_u16(&data, &mut index);
                    match code {

                        CODE_REGISTER => match PacketRegister::decode(&data, &mut index) {
                            Ok(packet) => {
                                if args.bus.send(bus_ds, QRPacket::new(conn_id, QRegister(packet))).is_ok() {
                                    while match args.bus.receive_wait() {
                                        Some(resp) => {
                                            let qr = &resp.data;
                                            match &qr.data {
                                                RRegister(resp) => {
                                                    ws.send(Message::Binary(
                                                        encode_response(code, resp.encode())
                                                    )).await.ok();
                                                    false
                                                }
                                                _ => true,
                                            }
                                        }
                                        None => true,
                                    } { }
                                }
                            }
                            Err(_) => { }
                        }

                        CODE_AUTH => match PacketAuth::decode(&data, &mut index) {
                            Ok(packet) => {
                                if args.bus.send(bus_ds, QRPacket::new(conn_id, QAuth(packet))).is_ok() {
                                    while match args.bus.receive_wait() {
                                        Some(resp) => {
                                            let qr = &resp.data;
                                            match &qr.data {
                                                RAuth(resp) => {
                                                    ws.send(Message::Binary(
                                                        encode_response(code, resp.encode())
                                                    )).await.ok();
                                                    false
                                                }
                                                _ => true,
                                            }
                                        }
                                        None => true,
                                    } { }
                                }
                            }
                            Err(_) => { }
                        }

                        CODE_DEAUTH => {
                            args.bus.send(bus_ds, QRPacket::new(conn_id, QDeauth)).ok();
                        }

                        _ => { }
                    }
                    true
                }
                _ => {
                    println!("notice: received unexpected websocket data.");
                    true
                }
            }
            Err(_) => false,
        }
        None => false,
    } { }

    args.bus.send(bus_ds, QRPacket::new(conn_id, QDisconn)).ok();
    ws.close(None).await.ok();
    Ok(())
}
